{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE DeriveGeneric #-}
{-# OPTIONS_GHC -Wno-deprecations #-}

module Gargantext.MicroServices.ReverseProxy (
  microServicesProxyApp
  ) where

import Prelude

import Conduit
import Data.ByteString qualified as B
import Data.ByteString.Builder
import Data.ByteString.Char8 qualified as C8
import Data.Conduit.List qualified as CC
import Data.String
import Data.Text qualified as T
import Data.Text.Encoding qualified as TE
import GHC.Generics
import Gargantext.API.Admin.EnvTypes
import Gargantext.Database.Prelude (hasConfig)
import Gargantext.Prelude
import Gargantext.Prelude.Config (gc_frame_write_url)
import Network.HTTP.ReverseProxy
import Network.HTTP.Types (hCacheControl, RequestHeaders, hReferer)
import Network.HTTP.Types.Header (hHost)
import Network.URI as URI
import Network.Wai (Request, rawPathInfo, requestHeaders)
import Servant
import Servant.Auth.Swagger ()
import Servant.Server.Generic

--
-- Types
--

data ReverseProxyAPI mode = ReverseProxyAPI
  { notesProxy    :: mode :- "notes-proxy" :> NamedRoutes NotesProxy
  , notesSocketIo :: mode :- "socket.io"   :> NamedRoutes SocketIOProxy
  , meEndpoint    :: mode :- "me"          :> Raw
  } deriving Generic

data NotesProxy mode = NotesProxy
  { noteProxyEp :: mode :- Capture "frameId" T.Text
                        :> Raw
  } deriving Generic

data SocketIOProxy mode = SocketIOProxy
  { socketIoEp  :: mode :- QueryParam "noteId" T.Text :> Raw
  } deriving Generic

--
-- Server
--

microServicesProxyApp :: Env -> Application
microServicesProxyApp env = genericServe (server env)

server :: Env -> ReverseProxyAPI AsServer
server env = ReverseProxyAPI {
    notesProxy    = notesProxyImplementation env
  , notesSocketIo = socketIOProxyImplementation env
  , meEndpoint    = forwardServer env
  }

notesProxyImplementation :: Env -> NotesProxy AsServer
notesProxyImplementation env = NotesProxy {
    noteProxyEp = \_frameId -> forwardServer env
  }

socketIOProxyImplementation :: Env -> SocketIOProxy AsServer
socketIOProxyImplementation env = SocketIOProxy {
    socketIoEp = \_noteId -> forwardServer env
  }

forwardServer :: Env -> ServerT Raw m
forwardServer env =
  Tagged $ waiProxyToSettings forwardRequest proxySettings (env ^. env_manager)
  where
    writeFrameURL :: T.Text
    writeFrameURL = env ^. hasConfig . gc_frame_write_url

    assetPath :: T.Text
    assetPath
      | snd forwardedHost /= 80
      = writeFrameURL <> ":" <> T.pack (Prelude.show $ snd forwardedHost)
      | otherwise
      = writeFrameURL

    removeProxyPath :: ByteString -> ByteString
    removeProxyPath = TE.encodeUtf8
                    . T.replace "/notes-proxy" ""
                    . TE.decodeUtf8

    proxySettings :: WaiProxySettings
    proxySettings = defaultWaiProxySettings { wpsLogRequest = \req -> traceShow req $ pure ()
                                            , wpsProcessBody =
                                                \_req _res -> Just replaceRelativeLinks
                                            }

    replaceRelativeLinks :: ConduitT B.ByteString (Flush Builder) IO ()
    replaceRelativeLinks = CC.map flushReplace
      where
        -- FIXME(adn) performance.
        flushReplace :: B.ByteString -> Flush Builder
        flushReplace bs = Chunk $ byteString $ traceShowId $ TE.encodeUtf8 . replaceIt . TE.decodeUtf8 $ bs

        replaceIt :: T.Text -> T.Text
        replaceIt t = t & T.replace "src=\"/build/" ("src=\"" <> assetPath <> "/build/")
                        & T.replace "href=\"/build/" ("href=\"" <> assetPath <> "/build/")
                        & T.replace "src=\"/config" ("src=\"" <> assetPath <> "/config")
                        & T.replace "src=\"/js" ("src=\"" <> assetPath <> "/js")

    noCache :: RequestHeaders -> RequestHeaders
    noCache hdrs = (hCacheControl, fromString "no-cache") :
      filter ((/=) hCacheControl . fst) hdrs

    newHost :: RequestHeaders -> RequestHeaders
    newHost hdrs =
      (hHost, fromString $ fst forwardedHost <> ":" <> Prelude.show (snd forwardedHost)) :
      filter ((/=) hHost . fst) hdrs

    newReferer :: RequestHeaders -> RequestHeaders
    newReferer hdrs =
      (hReferer, "http://localhost:8009") :
      filter ((/=) hHost . fst) hdrs

    forwardedHost :: (String, Int)
    forwardedHost = fromMaybe (panicTrace "Invalid URI found in the proxied Request.") $ do
      URI{..} <- URI.parseURI (T.unpack writeFrameURL)
      uriAut  <- uriAuthority
      port    <- case uriPort uriAut of
        "" -> pure 80
        xs -> readMaybe (dropWhile (==':') xs)
      pure $ (uriRegName uriAut, port)

    mkProxyDest :: ProxyDest
    mkProxyDest =
      let (hst, prt) = forwardedHost
      in ProxyDest (C8.pack hst) prt

    -- | Forwards the request by substituting back the proxied address into the actual one.
    forwardRequest :: Request -> IO WaiProxyResponse
    forwardRequest originalRequest = do
      let proxiedReq = originalRequest {
                           rawPathInfo = removeProxyPath (rawPathInfo originalRequest)
                         , requestHeaders = (newReferer . newHost . noCache $ (requestHeaders originalRequest))
                         }
      pure $ traceShow ("I'm PROXYING ===> " <> (Prelude.show originalRequest)) $ WPRModifiedRequest (traceShowId proxiedReq) $ traceShowId $ mkProxyDest
